; 
    include \masm32\include\masm32rt.inc
    .686
    include \masm32\macros\timers.asm

    ;----------------
    ; This by drizz.
    ;----------------

    AbsEAX macro
      cdq
      xor eax,edx
      sub eax,edx
    endm

; 
    .data
      error     dq 0
      sqrt_tbl  dd 0
      sqrt      dd 0
    .code
; 

align 4
uddsqrt proc uses esi edi ebx ecx edx number: dword

  mov ebx, number  ; get initial copy of number
  xor edx, edx     ; mov rem, 0
  xor eax, eax     ; mov root, 0
 ;   xor esi, esi     ; esi = 0
 ;   mov  ch, 16      ; mov i, 16           16 12  8  4  iterations  half the number of bits
 ;   mov  cl, 30      ; 2 less than         30 22 14  6  right shifts to get the top 2 bits
 ;   dec esi          ; esi = -1            32 24 16  8  bits set for ANDing

  .if ebx <= 255              ; fits in  8 bits
     mov  ch,  4   ; iterations for  8 bit
     mov  cl,  6   ; shifts needed
     mov esi, 0ffh
  .elseif ebx <= 65535            ; fits in 16 bits
     mov  ch,  8   ; iterations for 16 bit
     mov  cl, 14   ; shifts needed
     mov esi, 0ffffh
  .elseif ebx <= 65536*256 - 1    ; fits in 24 bits
     mov  ch, 12   ; iterations for 24 bit
     mov  cl, 22   ; shifts needed
     mov esi, 0ffffffh
  .else
     mov  ch, 16   ; iterations for 32 bit
     mov  cl, 30   ; shifts needed
     xor esi, esi
     dec esi
  .endif   

align 16
@@:
  add eax, eax    ; shl root, 1       shl eax, 1
  shl edx, 2      ; shl rem, 2

 ;   original version of get top two bits of current number
  mov edi, ebx    ; mov temp, number
  shr edi,  cl    ; shr temp, (one of 30, 22, 14, 6) 

 ;   new version turns out slower, not as flexible
 ;   shld edi, ebx, 2  ; get top two bits of current number  (only works for 32 bit)

  shl ebx, 2      ; shl number, 2
  add edx, edi    ; add rem, temp
  and ebx, esi    ; keep number within bit fit limit

 ;   original version of subtract root from rem
 ;   inc eax         ; inc root
 ;   .if eax <= edx  ; .if root <= rem
 ;      sub edx, eax ;    sub rem, root
 ;      inc eax      ;    inc root
 ;   .else
 ;      dec eax      ;    dec root  undo inc 
 ;   .endif

 ;   new version vastly faster 
  .if eax < edx      ; .if root < rem
     inc eax         ;    inc root
     sub edx, eax    ;    sub rem, root
     inc eax         ;    inc root
  .endif             ; .endif

  dec  ch         ; dec i
  jnz @B
  shr eax, 1      ; shr root, 1
  ret
uddsqrt endp

; 
align 4
fisqrt proc arg:DWORD
    fild  arg
    fsqrt
    push  eax
    fistp DWORD PTR [esp]
    pop   eax
    ret
fisqrt endp

; 
start:
; 

    xor ebx, ebx
    .WHILE ebx < 100
      invoke nrandom,1000
      mov esi, eax
      print ustr$(esi)
      loc 15, ebx
      invoke uddsqrt, esi
      print ustr$(eax)
      loc 25, ebx
      invoke fisqrt, esi
      print ustr$(eax),13,10
      inc ebx
    .ENDW

    .WHILE ebx < 200
      invoke nrandom,4000000000
      mov esi, eax
      print ustr$(esi)
      loc 15, ebx
      invoke uddsqrt, esi
      print ustr$(eax)
      loc 25, ebx
      invoke fisqrt, esi
      print ustr$(eax),13,10
      inc ebx
    .ENDW

    print chr$(13,10)
    inkey
    print chr$(13,10)

    xor esi, esi
    xor ebx, ebx
    .WHILE ebx < 1000
      invoke uddsqrt, ebx
      push eax
      invoke fisqrt, ebx
      pop edx
      sub eax, edx
      AbsEAX
      add esi, eax
      inc ebx
    .ENDW
    push esi
    fild DWORD PTR [esp]
    fld8 1000.0
    fdiv
    fstp error
    pop eax
    invoke crt_printf, chr$("mean absolute error 0-999: %f%c%c"),
                       error, 10, 10

    xor esi, esi
    xor ebx, ebx
    .WHILE ebx < 1000000
      invoke uddsqrt, ebx
      push eax
      invoke fisqrt, ebx
      pop edx
      sub eax, edx
      AbsEAX
      add esi, eax
      inc ebx
    .ENDW
    push esi
    fild DWORD PTR [esp]
    fld8 1000000.0
    fdiv
    fstp error
    pop eax
    invoke crt_printf, chr$("mean absolute error 0-999999: %f%c%c"),
                       error, 10, 10

    xor esi, esi
    xor ebx, ebx
    .WHILE ebx < 10000000
      invoke uddsqrt, ebx
      push eax
      invoke fisqrt, ebx
      pop edx
      sub eax, edx
      AbsEAX
      add esi, eax
      inc ebx
    .ENDW
    push esi
    fild DWORD PTR [esp]
    fld8 10000000.0
    fdiv
    fstp error
    pop eax
    invoke crt_printf, chr$("mean absolute error 0-9999999: %f%c%c"),
                       error, 10, 10

    xor esi, esi
    xor ebx, ebx
    .WHILE ebx < 100000000
      invoke uddsqrt, ebx
      push eax
      invoke fisqrt, ebx
      pop edx
      sub eax, edx
      AbsEAX
      add esi, eax
      inc ebx
    .ENDW
    push esi
    fild DWORD PTR [esp]
    fld8 100000000.0
    fdiv
    fstp error
    pop eax
    invoke crt_printf, chr$("mean absolute error 0-99999999: %f%c%c"),
                       error, 10, 10

    inkey
    print chr$(13,10)

    counter_begin 1000, HIGH_PRIORITY_CLASS
      invoke uddsqrt, 123456
    counter_end
    print ustr$(eax)," cycles, uddsqrt",13,10

    counter_begin 1000, HIGH_PRIORITY_CLASS
      invoke fisqrt, 123456
    counter_end
    print ustr$(eax)," cycles, fisqrt",13,10,13,10

    ; free sqrt_tbl
    inkey "Press any key to exit..."
    exit
; 
end start
 